use super::{addr, Addr, Signature};
use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};
use serde_json::json;
use sha2::{Digest, Sha256};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Error serializing object: {0}")]
    Serialize(#[from] serde_json::Error),

    #[error("Error serializing id: {0}")]
    Id(#[from] addr::Error),
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct EventContent {
    pub pub_key: Addr,
    #[serde(with = "super::ts_seconds")]
    pub created_at: DateTime<Utc>,
    pub kind: u32,
    pub tags: Vec<super::Tag>,
    pub content: String,
}

impl EventContent {
    pub fn new<T>(
        pub_key: Addr,
        kind: u32,
        tags: Vec<super::Tag>,
        content_obj: &T,
        created_at: Option<DateTime<Utc>>,
    ) -> Result<Self, Error>
    where
        T: ?Sized + Serialize,
    {
        Ok(Self {
            pub_key,
            created_at: created_at.unwrap_or_else(|| Utc::now()),
            kind,
            tags,
            content: serde_json::to_string(&content_obj)?,
        })
    }

    pub fn id(&self) -> Result<Addr, Error> {
        let data_to_hash = serde_json::to_string(&json!(vec![
            json!(0),
            json!(self.pub_key.to_public_key()),
            json!(self.created_at.timestamp()),
            json!(self.kind),
            json!(self.tags),
            json!(self.content),
        ]))?;
        let mut hasher = Sha256::new();
        hasher.update(&data_to_hash);
        let id: [u8; 32] = hasher.finalize().into();
        Ok(Addr::try_from_bytes(Default::default(), id.to_vec())?)
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Event {
    pub id: Addr,
    #[serde(flatten)]
    pub content: EventContent,
    pub signature: Signature,
}

impl Event {
    pub fn new(content: EventContent, signature: Signature) -> Result<Self, Error> {
        Ok(Self {
            id: content.id()?,
            content,
            signature,
        })
    }
}
